!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_convert
!  NAME
!    cadna_convert
!
!
!  DESCRIPTION
!    Contains the functions and subroutines related to the conversion
!    between stochastic types and classical types, the overlaoding of
!    the classical internal conversion function of the FORTRAN 77 
!    and functions which allow to take into account data uncertainty
!    at initialisation of stochastic variables.
!
!******
!
module cadna_convert
  !
  use cadna_type
  !
!****f cadna_convert/old_type
!  NAME
!    old_type
!
!  FUNCTION
!    Functions of the interface "old_type" convert a stochastic type to
!    the corresponding classical type.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A classical REAL or DOUBLE PRECISION value.
!
!
!******
  interface old_type
     !       ======================= 
     module procedure old_st, old_dst
  end interface
  !
!****f cadna_convert/int
!  NAME
!    int
!
!  FUNCTION
!    Functions of the interface "int" are the equivalent of classical
!    INT function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A INTEGER value.
!
!
!******
  interface int
     !       ======================= 
     module procedure int_st, int_dst
  end interface
  !
!****f cadna_convert/nint
!  NAME
!    nint
!
!  FUNCTION
!    Functions of the interface "nint" are the equivalent of classical
!    NINT function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    A INTEGER value.
!
!
!******
  interface nint
     !       ======================= 
     module procedure nint_st, nint_dst
  end interface
  !
!****f cadna_convert/aint
!  NAME
!    aint
!
!  FUNCTION
!    Functions of the interface "aint" are the equivalent of classical
!    AINT function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    A value of the same type as the argument.
!
!
!******
  interface aint
     !       ======================= 
     module procedure aint_st, aint_dst
  end interface
  !
  interface anint
     !       ======================= 
!****f cadna_convert/anint
!  NAME
!    anint
!
!  FUNCTION
!    Functions of the interface "anint" are the equivalent of classical
!    ANINT function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A value of the same type as the argument.
!
!
!******
     module procedure anint_st, anint_dst
  end interface
  !
!****f cadna_convert/real
!  NAME
!    real
!
!  FUNCTION
!    Functions of the interface "real" are the equivalent of classical
!    REAL function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A value of SINGLE_ST type.
!
!
!******
  interface real
     !       ======================= 
     module procedure real_st, real_dst
  end interface
  !
!****f cadna_convert/dble
!  NAME
!    dble
!
!  FUNCTION
!    Functions of the interface "dble" are the equivalent of classical
!    DBLE function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A value of DOUBLE_ST type.
!
!
!******
  interface dble
     !       ======================= 
     module procedure dble_st, dble_dst
  end interface
  !
!****f cadna_convert/data_st
!  NAME
!    data_st
!
!  FUNCTION
!    Subroutines of the interface "data_st" allow to take into account
!    data uncertainty at the initialization of stochastic variables.
!
!  INPUTS
!    a: single_st (or double_st), an in/out argument on which data
!    uncertainty will be applied.
!
!    e: real (or double precision), an optional argument which is 
!    the relative or absolute uncertainty of the previous argument.
!    if this argument is not present, only perturbation on the last
!    bit of the mantissa is performed. 
!
!    ierr: integer, an optional argument. If IERR is 1, uncertainty 
!    is relative and absolute for an other value. If it is not present,
!    it is the same as IERR = 1.
!
!  OUTPUTS
!    The same "input" variable "a".
!
!******
  interface data_st
     !       ======================= 
     module procedure data_s_st, data_s_dst, &
          &           data_e_st, data_e_dst
  end interface
  !
!****f cadna_convert/sngl
!  NAME
!    sngl
!
!  FUNCTION
!    Functions of the interface "sngl" are the equivalent of classical
!    SNGL function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A value of SINGLE_ST type.
!
!
!******
  interface sngl
     !       ======================= 
     module procedure sngl_dst
  end interface
  !
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  real function old_st(a)
    implicit none
    type (single_st), intent(in):: a
    old_st = ( a%x + a%y + a%z) / 3.0
  end function old_st
  !
  double precision function old_dst(a)
    implicit none
    type (double_st), intent(in):: a
    old_dst = ( a%x + a%y + a%z) / 3.0
  end function old_dst
  !
  integer function int_st(a)
    implicit none
    type (single_st), intent(in):: a
    if (cadna_intrinsic_tag) then
       if (int(a%x).ne.int(a%y).or.int(a%x).ne.int(a%z)) then
          call instability(cadna_intrinsic)
       end if
    endif
    int_st = int(( a%x + a%y + a%z ) / 3.0)
  end function int_st
  !
  integer function int_dst(a)
    implicit none
    type (double_st), intent(in):: a
    if (cadna_intrinsic_tag) then
       if (int(a%x).ne.int(a%y).or.int(a%x).ne.int(a%z)) then
          call instability(cadna_intrinsic)
       end if
    endif
    int_dst = int(( a%x + a%y + a%z ) / 3.0)
  end function int_dst
!
  integer function nint_st(a)
    implicit none
    type (single_st), intent(in):: a
    if (cadna_intrinsic_tag) then
       if (nint(a%x).ne.nint(a%y).or.nint(a%x).ne.nint(a%z)) then
          call instability(cadna_intrinsic)
       end if
    endif
    nint_st = nint(( a%x + a%y + a%z ) / 3.0)
  end function nint_st
  !
  integer function nint_dst(a)
    implicit none
    type (double_st), intent(in):: a
    if (cadna_intrinsic_tag) then
       if (nint(a%x).ne.nint(a%y).or.nint(a%x).ne.nint(a%z)) then
          call instability(cadna_intrinsic)
       end if
    endif
    nint_dst = nint(( a%x + a%y + a%z ) / 3.0)
  end function nint_dst
  !
  type (single_st) function aint_st(a)
    implicit none
    type (single_st), intent(in) :: a
    aint_st%x = aint(a%x)
    aint_st%y = aint(a%y)
    aint_st%z = aint(a%z)
    aint_st%acc = -1
  end function aint_st
  !
  type (double_st) function aint_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    aint_dst%x = aint(a%x)
    aint_dst%y = aint(a%y)
    aint_dst%z = aint(a%z)
    aint_dst%acc = -1
  end function aint_dst
  !
  type (single_st) function anint_st(a)
    implicit none
    type (single_st), intent(in) :: a
    anint_st%x = anint(a%x)
    anint_st%y = anint(a%y)
    anint_st%z = anint(a%z)
    anint_st%acc = -1
  end function anint_st
  !
  type (double_st) function anint_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    anint_dst%x = anint(a%x)
    anint_dst%y = anint(a%y)
    anint_dst%z = anint(a%z)
    anint_dst%acc = -1
  end function anint_dst
  !
  type (single_st) function real_st(a)
    implicit none
    type (single_st), intent(in) :: a
    real_st = a
  end function real_st
  !
  type (single_st) function real_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    real_dst%x = a%x
    real_dst%y = a%y
    real_dst%z = a%z
    real_dst%acc = min(a%acc,7)
  end function real_dst
  !
  type (double_st) function dble_st(a)
    implicit none
    type (single_st), intent(in) :: a
    dble_st%x = a%x
    dble_st%y = a%y
    dble_st%z = a%z
    dble_st%acc = a%acc
  end function dble_st
  !
  type (double_st) function dble_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    dble_dst = a
  end function dble_dst
  !
  subroutine data_s_st(a)
    implicit none
    type (single_st), intent(inout) :: a
    real :: aux=1.0
    logical :: b
    if (a%x.eq.a%y) then
       if (aux+1.0e-20.eq.1.0) then
          b = .true.
       else
          b = .false.
       endif
       call rnd_arr
       a%x = nearest(a%x,1.0)
       a%y = nearest(a%y,-1.0)
       if (b) then
          call rnd_moinf
       else
          call rnd_plinf
       endif
    endif
  end subroutine data_s_st
  !
  subroutine data_s_dst(a)
    implicit none
    type (double_st), intent(inout) :: a
    real :: aux=1.0
    logical :: b
    if (a%x.eq.a%y) then
       if (aux+1.0e-20.eq.1.0) then
          b = .true.
       else
          b = .false.
       endif
       call rnd_arr
       a%x = nearest(a%x,1.0)
       a%y = nearest(a%y,-1.0)
       if (b) then
          call rnd_moinf
       else
          call rnd_plinf
       endif
    endif
  end subroutine data_s_dst
  !
  subroutine data_e_st(a,e,ierr)
    implicit none
    type (single_st), intent(inout) :: a
    integer, optional :: ierr
    integer :: mierr
    real :: e,aux
    real :: aux1=1.0
    logical :: b
    if (aux1+1.0e-20.eq.1.0) then
       b = .true.
    else
       b = .false.
    endif
    if (present(ierr)) then
       mierr = ierr
    else
       mierr = 0
    endif
    if (mierr .ne. 1) then
       call rnd_plinf
       call random_number(aux)
       a%x = a%x*(1.0 + aux*e)
       call rnd_moinf
       call random_number(aux)
       a%y = a%y*(1.0 - aux*e)
       a%acc = -1
    else 
       call rnd_plinf
       call random_number(aux)
       a%x = a%x + aux*e
       call rnd_moinf
       call random_number(aux)
       a%y = a%y - aux*e
       a%acc = -1
    endif
    if (b) then	
       call rnd_moinf
    else
       call rnd_plinf
    endif
  end subroutine data_e_st
  !
  subroutine data_e_dst(a,e,ierr)
    implicit none
    type (double_st), intent(inout) :: a
    integer, optional :: ierr
    integer :: mierr
    double precision e,aux
    real :: aux1=1.0
    logical :: b
    if (aux1+1.0e-20.eq.1.0) then
       b = .true.
    else
       b = .false.
    endif
    if (present(ierr)) then
       mierr = ierr
    else
       mierr = 0
    endif
    if (mierr .ne. 1) then
       call rnd_plinf
       call random_number(aux)
       a%x = a%x*(1.0 + aux*e)
       call rnd_moinf
       call random_number(aux)
       a%y = a%y*(1.0 - aux*e)
       a%acc = -1
    else 
       call rnd_plinf
       call random_number(aux)
       a%x = a%x + aux*e
       call rnd_moinf
       call random_number(aux)
       a%y = a%y - aux*e
       a%acc = -1
    endif
    if (b) then	
       call rnd_moinf
    else
       call rnd_plinf
    endif
  end subroutine data_e_dst
  !
  type (single_st) function sngl_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    sngl_dst=real_dst(a)
  end function sngl_dst
  ! 
end module cadna_convert
