!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_equ
!  NAME
!    cadna_equ
!
!
!  DESCRIPTION
!    Defines all the functions involving at least one argument
!    of stochastic type which overload the "==" and "/=" 
!    operators in a test such as "a==b" or "a/=b".
!    
!  INPUTS
!    All the functions have two arguments of stochastic, 
!    real or integer type in single or double precision. At least 
!    one argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one logical return value. 
!
!  NOTES
!    The return value takes into account the stochastic definition of
!    the "==" operator (see the documentation). With this overloading,
!    the classical ".eq." and ".ne." operators are still available for
!    stochastic variables.
!
!******
!
module cadna_equ
  !
  use cadna_type
  use cadna_cestac
  !
  !
  interface operator (==)
     !	=======================
     module procedure equ_st_st, &
          &		equ_single_st, equ_st_single, equ_int_st, equ_st_int, &
          &		equ_dst_dst, equ_dst_double, equ_double_dst, &
          &		equ_dst_st, equ_st_dst, equ_dst_single, equ_single_dst, &
          &		equ_st_double, equ_double_st, equ_int_dst, equ_dst_int
  end interface
  !
  interface operator (.ne.)
     !	=======================
     module procedure ne_st_st, &
          &		ne_single_st, ne_st_single, ne_int_st, ne_st_int, &
          &		ne_dst_dst, ne_dst_double, ne_double_dst, &
          &		ne_dst_st, ne_st_dst, ne_dst_single, ne_single_dst, &
          &		ne_st_double, ne_double_st, ne_int_dst, ne_dst_int
  end interface
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  logical function equ_st_st(a,b)
    implicit none
    type  (single_st), intent(in) :: a,b
    type (single_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    equ_st_st = computed_zero_st(aux)
  end function equ_st_st
  ! 
  logical function equ_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type  (single_st), intent(in) :: b
    type (single_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    equ_single_st = computed_zero_st(aux)
  end function equ_single_st
  ! 
  logical function equ_st_single(a,b)
    implicit none
    type  (single_st), intent(in) :: a
    real, intent(in) :: b
    type (single_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    equ_st_single = computed_zero_st(aux)
  end function equ_st_single
  ! 
  logical function equ_int_st(a,b)
    implicit none
    integer, intent(in) :: a
    type  (single_st), intent(in) :: b
    type (single_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    equ_int_st = computed_zero_st(aux)
  end function equ_int_st
  ! 
  logical function equ_st_int(a,b)
    implicit none
    type  (single_st), intent(in) :: a
    integer, intent(in) :: b
    type (single_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    equ_st_int = computed_zero_st(aux)
  end function equ_st_int
  !
  logical function equ_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a,b
    type (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    equ_dst_dst = computed_zero_dst(aux)
  end function equ_dst_dst
  !
  logical function equ_dst_double(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    equ_dst_double = computed_zero_dst(aux)
  end function equ_dst_double
  !
  logical function equ_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    equ_double_dst = computed_zero_dst(aux)
  end function equ_double_dst
  !
  logical function equ_dst_st(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    equ_dst_st = computed_zero_dst(aux)
  end function equ_dst_st
  !
  logical function equ_st_dst(a,b)
    implicit none
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    equ_st_dst = computed_zero_dst(aux)
  end function equ_st_dst
  !
  logical function equ_single_dst(a,b)
    implicit none
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    equ_single_dst = computed_zero_dst(aux)
  end function equ_single_dst
  !
  logical function equ_dst_single(a,b)
    implicit none
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    equ_dst_single = computed_zero_dst(aux)
  end function equ_dst_single
  !
  logical function equ_st_double(a,b)
    implicit none
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    equ_st_double = computed_zero_dst(aux)
  end function equ_st_double
  !
  logical function equ_double_st(a,b)
    implicit none
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    equ_double_st = computed_zero_dst(aux)
  end function equ_double_st
  ! 
  logical function equ_int_dst(a,b)
    implicit none
    integer, intent(in) :: a
    type  (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    equ_int_dst = computed_zero_dst(aux)
  end function equ_int_dst
  ! 
  logical function equ_dst_int(a,b)
    implicit none
    type  (double_st), intent(in) :: a
    integer, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    equ_dst_int = computed_zero_dst(aux)
  end function equ_dst_int
  !
  logical function ne_st_st(a,b)
    implicit none
    type  (single_st), intent(in) :: a,b
    type (single_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    ne_st_st = .not.computed_zero_st(aux)
  end function ne_st_st
  ! 
  logical function ne_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type  (single_st), intent(in) :: b
    type (single_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    ne_single_st = .not.computed_zero_st(aux)
  end function ne_single_st
  ! 
  logical function ne_st_single(a,b)
    implicit none
    type  (single_st), intent(in) :: a
    real, intent(in) :: b
    type (single_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    ne_st_single = .not.computed_zero_st(aux)
  end function ne_st_single
  ! 
  logical function ne_int_st(a,b)
    implicit none
    integer, intent(in) :: a
    type  (single_st), intent(in) :: b
    type (single_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    ne_int_st = .not.computed_zero_st(aux)
  end function ne_int_st
  ! 
  logical function ne_st_int(a,b)
    implicit none
    type  (single_st), intent(in) :: a
    integer, intent(in) :: b
    type (single_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    ne_st_int = .not.computed_zero_st(aux)
  end function ne_st_int
  !
  logical function ne_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a,b
    type (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    ne_dst_dst = .not.computed_zero_dst(aux)
  end function ne_dst_dst
  !
  logical function ne_dst_double(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    ne_dst_double = .not.computed_zero_dst(aux)
  end function ne_dst_double
  !
  logical function ne_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    ne_double_dst = .not.computed_zero_dst(aux)
  end function ne_double_dst
  !
  logical function ne_dst_st(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    ne_dst_st = .not.computed_zero_dst(aux)
  end function ne_dst_st
  !
  logical function ne_st_dst(a,b)
    implicit none
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    ne_st_dst = .not.computed_zero_dst(aux)
  end function ne_st_dst
  !
  logical function ne_single_dst(a,b)
    implicit none
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    ne_single_dst = .not.computed_zero_dst(aux)
  end function ne_single_dst
  !
  logical function ne_dst_single(a,b)
    implicit none
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    ne_dst_single = .not.computed_zero_dst(aux)
  end function ne_dst_single
  !
  logical function ne_st_double(a,b)
    implicit none
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    ne_st_double = .not.computed_zero_dst(aux)
  end function ne_st_double
  !
  logical function ne_double_st(a,b)
    implicit none
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    ne_double_st = .not.computed_zero_dst(aux)
  end function ne_double_st
  ! 
  logical function ne_int_dst(a,b)
    implicit none
    integer, intent(in) :: a
    type  (double_st), intent(in) :: b
    type (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    ne_int_dst = .not.computed_zero_dst(aux)
  end function ne_int_dst
  ! 
  logical function ne_dst_int(a,b)
    implicit none
    type  (double_st), intent(in) :: a
    integer, intent(in) :: b
    type (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    ne_dst_int = .not.computed_zero_dst(aux)
  end function ne_dst_int
  !
end module cadna_equ
