!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_intr
!  NAME
!    cadna_intr
!
!
!  DESCRIPTION
!    Contains the overloading of the classical internal functions 
!    in FORTRAN 77: ABS, SIGN, MAX, MIN, MOD, DIM.
!
!******
!
module cadna_intr
  !
  use cadna_type
  use cadna_cestac
!  use cadna_to
  !
!****f cadna_intr/abs
!  NAME
!    abs
!
!  FUNCTION
!    Functions of the interface "abs" are the equivalent of classical
!    ABS function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!    a can be an array of rank 1 in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface abs
     !       ======================= 
     module procedure abs_st, abs_dst, abs_vdst, abs_vvdst, abs_vvvdst
  end interface
  !
!****f cadna_intr/sign
!  NAME
!    sign
!
!  FUNCTION
!    Functions of the interface "sign" are the equivalent of classical
!    SIGN function for stochastic types.
!
!  INPUTS
!    All the functions have two arguments of stochastic or real type in 
!    single or double precision. At least one argument must be of 
!    stochastic type. The two argument must be in the same precision.
!
!  RETURN VALUE
!    A numerical stochastic variable in the same precision as the 
!    input argument. 
!
!
!******
  interface sign
     !       ======================
     module procedure sign_st_st, &
          &		sign_single_st, sign_st_single, &
          &		sign_dst_dst, sign_dst_double, &
          &		sign_double_dst, sign_double_vvdst
  end interface
  !
!****f cadna_intr/max
!  NAME
!    max
!
!  FUNCTION
!    Functions of the interface "max" are the equivalent of classical
!    MAX function for stochastic types.
!
!  INPUTS
!    All the functions have two arguments of stochastic or real type in 
!    single or double precision. At least one argument must be of 
!    stochastic type. The two argument must be in the same precision.
!    Argument can be an array of rank 1 only in double precision.
!
!  RETURN VALUE
!    A numerical stochastic variable in the same precision as the 
!    input argument. It is an array of rank 1 if one input argument 
!    is an array of rank 1.
!
!
!******
  interface max
     !       ======================
     module procedure &
          &          max_st_st, max_single_st, max_st_single, &
          &		max_dst_dst, max_dst_double, max_double_dst, &
          !-----------------------------------------------------------
          &          max_vdst_vdst, max_vdst_vdouble, max_vdouble_vdst, max_dst_vdst, max_vdst_dst, &
          &          max_vdst_double, max_double_vdst, max_vdouble_dst, max_dst_vdouble, &
          !-----------------------------------------------------------
          &          max_dst_dst_dst, &
          &          max_dst_dst_dst_dst, &
          &          max_dst_dst_dst_dst_dst, &
          &          max_dst_dst_dst_dst_dst_dst, &
          &          max_dst_dst_dst_dst_dst_dst_dst, &
          &          max_dst_dst_dst_dst_dst_dst_dst_dst, &
          &          max_dst_dst_dst_dst_dst_dst_dst_dst_dst, &
          &          max_dst_dst_double, &
          &          max_double_dst_dst_dst, &
          &          max_vvvdst_double, max_double_vvvdst, &
          &          max_vvvdst_vvvdst
  end interface
  !
!****f cadna_intr/min
!  NAME
!    min
!
!  FUNCTION
!    Functions of the interface "min" are the equivalent of classical
!    MIN function for stochastic types.
!
!  INPUTS
!    All the functions have two arguments of stochastic or real type in 
!    single or double precision. At least one argument must be of 
!    stochastic type. The two argument must be in the same precision.
!    Argument can be an array of rank 1 only in double precision.
!
!  RETURN VALUE
!    A numerical stochastic variable in the same precision as the 
!    input argument. It is an array of rank 1 if one input argument 
!    is an array of rank 1.
!
!
!******
  interface min
     !       ======================
     module procedure &
          &          min_st_st, min_single_st, min_st_single, &
          &          min_dst_dst, min_dst_double, min_double_dst, &
          !-----------------------------------------------------------
          &          min_vdst_vdst, min_vdst_vdouble, min_vdouble_vdst, min_dst_vdst, min_vdst_dst, &
          &          min_vdst_double, min_double_vdst, min_vdouble_dst, min_dst_vdouble, &
          !-----------------------------------------------------------
          &          min_dst_dst_dst, &
          &          min_dst_dst_dst_dst, &
          &          min_dst_dst_dst_dst_dst, &
          &          min_dst_dst_dst_dst_dst_dst, &
          &          min_dst_dst_dst_dst_dst_dst_dst, &
          &          min_dst_dst_dst_dst_dst_dst_dst_dst, &
          &          min_dst_dst_dst_dst_dst_dst_dst_dst_dst, &
          &          min_dst_dst_double, &
          &          min_vvvdst_double, min_double_vvvdst, &
          &          min_vvvdst_vvvdst
  end interface
  !
!****f cadna_intr/mod
!  NAME
!    mod
!
!  FUNCTION
!    Functions of the interface "mod" are the equivalent of classical
!    MOD function for stochastic types.
!
!  INPUTS
!    All the functions have two arguments of stochastic or real type in 
!    single or double precision. At least one argument must be of 
!    stochastic type. The two argument must be in the same precision.
!
!  RETURN VALUE
!    A numerical stochastic variable in the same precision as the 
!    input argument. 
!
!
!******
  interface mod
     !       ======================
     module procedure mod_st_st, &
          &		mod_single_st, mod_st_single, &
          &		mod_dst_dst, mod_dst_double, &
          &		mod_double_dst
  end interface
  interface modulo
     !       ======================
     module procedure modulo_dst_double
  end interface
  !
!****f cadna_intr/dim
!  NAME
!    dim
!
!  FUNCTION
!    Functions of the interface "dim" are the equivalent of classical
!    DIM function for stochastic types.
!
!  INPUTS
!    All the functions have two arguments of stochastic or real type in 
!    single or double precision. At least one argument must be of 
!    stochastic type. The two argument must be in the same precision.
!
!  RETURN VALUE
!    A numerical stochastic variable in the same precision as the 
!    input argument. 
!
!
!******
  interface dim
     !       ======================
     module procedure dim_st_st, &
          &		dim_single_st, dim_st_single, &
          &		dim_dst_dst, dim_dst_double, &
          &		dim_double_dst
  end interface
  !
!****f cadna_intr/minval
!  NAME
!    minval
!
!  FUNCTION
!    Functions of the interface "minval" are the equivalent of classical
!    MINVAL function for stochastic types.
!
!  INPUTS
!    a: double_st array
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface minval
     !       ======================= 
     module procedure minval_vdst, minval_vvdst, minval_vvdst_dim
  end interface
  !
!****f cadna_intr/maxval
!  NAME
!    maxval
!
!  FUNCTION
!    Functions of the interface "maxval" are the equivalent of classical
!    MAXVAL function for stochastic types.
!
!  INPUTS
!    a: double_st array
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface maxval
     !       ======================= 
     module procedure maxval_vdst, maxval_vvdst, maxval_vvdst_dim, maxval_vvvdst
  end interface
  !
!****f cadna_intr/matmul
!  NAME
!    matmul
!
!  FUNCTION
!    Functions of the interface "matmul" are the equivalent of classical
!    MATMUL function for stochastic types.
!
!  INPUTS
!    a, b: double_st arrays
!
!  RETURN VALUE
!    A numerical array of the same type as the input arguments. 
!
!
!******
  interface matmul
     !       ======================= 
     module procedure matmul_vvdst_vdst, matmul_vdst_vvdst, matmul_vvdst_vvdst
  end interface
  !
!****f cadna_intr/dot_product
!  NAME
!    dot_product
!
!  FUNCTION
!    Functions of the interface "dot_product" are the equivalent of classical
!    DOT_PRODUCT function for stochastic types.
!
!  INPUTS
!    a, b: double_st arrays
!
!  RETURN VALUE
!    A numerical array of the same type as the input argument. 
!
!
!******
  interface dot_product
     !       ======================= 
     module procedure dot_product_vdst_vdst
  end interface
  !
!****f cadna_intr/huge
!  NAME
!    huge
!
!  FUNCTION
!    Functions of the interface "huge" are the equivalent of classical
!    HUGE function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    A double (or single) précision numerical variable. 
!
!
!******
  interface huge
     !       ======================= 
     module procedure huge_dst
  end interface
  !
!****f cadna_intr/sum
!  NAME
!    sum
!
!  FUNCTION
!    Functions of the interface "sum" are the equivalent of classical
!    SUM function for stochastic types.
!
!  INPUTS
!    xxx
!
!  RETURN VALUE
!    A numerical stochastic variable in the same precision as the 
!    input argument.
!
!
!******
  interface sum
     !       ======================
     module procedure &
          &          sum_dst_vdst
  end interface
  !
!****f cadna_intr/floor
!  NAME
!    floor
!
!  FUNCTION
!    Functions of the interface "floor" are the equivalent of classical
!    FLOOR function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!    a can be an array of rank 1 in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface floor
     !       ======================= 
     module procedure floor_dst
  end interface
  !
!****f cadna_intr/ceiling
!  NAME
!    ceiling
!
!  FUNCTION
!    Functions of the interface "ceiling" are the equivalent of classical
!    CEILING function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!    a can be an array of rank 1 in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface ceiling
     !       ======================= 
     module procedure ceiling_dst
  end interface
  !
!****f cadna_intr/cpu_time
!  NAME
!    cpu_time
!
!  FUNCTION
!    Functions of the interface "cpu_time" are the equivalent of classical
!    CPU_TIME function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!    a can be an array of rank 1 in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface cpu_time
     !       ======================= 
     module procedure cpu_time_dst
  end interface
  !
!****f cadna_intr/epsilon
!  NAME
!    epsilon
!
!  FUNCTION
!    Functions of the interface "epsilon" are the equivalent of classical
!    EPSILON function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!    a can be an array of rank 1 in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface epsilon
     !       ======================= 
     module procedure epsilon_dst
  end interface
  !
!****f cadna_intr/nint
!  NAME
!    nint
!
!  FUNCTION
!    Functions of the interface "nint" are the equivalent of classical
!    NINT function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!    a can be an array of rank 1 in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!
!******
  interface nint
     !       ======================= 
     module procedure nint_vdst, nint_vvdst
  end interface
  !
!****f cadna_intr/maxloc
!  NAME
!    maxloc
!
!  FUNCTION
!    Functions of the interface "maxloc" are the equivalent of the 
! classical MAXLOC function for stochastic types.
!
!  INPUTS
!    All the functions have one array argument of stochastic type in 
!    double precision. 
!
!  RETURN VALUE
!    The result is of integer type. 
!
!
!******
  interface maxloc
     !       ======================
     module procedure maxloc_vdst, maxloc_vvdst
  end interface
  !
!****f cadna_intr/minloc
!  NAME
!    minloc
!
!  FUNCTION
!    Functions of the interface "minloc" are the equivalent of the
!    classical MINLOC function for stochastic types.
!
!  INPUTS
!    All the functions have one array argument of stochastic type in 
!    double precision. 
!
!  RETURN VALUE
!    The result is of integer type. 
!
!
!******
  interface minloc
     !       ======================
     module procedure minloc_vdst, minloc_vvdst
  end interface
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  type (single_st) function abs_st(a)
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_intrinsic_tag) then
       if (a%acc.eq.-1) call accuracy_st(a) 
       if (a%acc.eq.0) then
          call instability(cadna_intrinsic)
       end if
    end if
    if (a%x.ge.0.0) then
       abs_st = a
    else
       abs_st%x = abs(a%x)
       abs_st%y = abs(a%y)
       abs_st%z = abs(a%z)
       abs_st%acc = a%acc
    end if
  end function abs_st
  !
  type (double_st) function abs_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_intrinsic_tag) then
       if (a%acc.eq.-1) call accuracy_dst(a) 
       if (a%acc.eq.0) then
          call instability(cadna_intrinsic)
       end if
    end if
    if (a%x.ge.0.0) then
       abs_dst = a
    else
       abs_dst%x = abs(a%x)
       abs_dst%y = abs(a%y)
       abs_dst%z = abs(a%z)
       abs_dst%acc = a%acc
    end if
  end function abs_dst
  !
  !
!  function abs_vst(a)
!    implicit none
!    type (single_st), dimension(:), intent(in) :: a
!    type (single_st), dimension(size(a)) :: abs_vst
!    integer :: i
!    do i=1,size(a)
!       abs_vst(i) = abs_st(a(i))
!    enddo
!  end function abs_vst
  !
  function abs_vdst(a)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: abs_vdst
    integer :: i
    do i=1,size(a)
       abs_vdst(i) = abs_dst(a(i))
    enddo
  end function abs_vdst
  !
  function abs_vvdst(a)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1),size(a,2)) :: abs_vvdst
    integer :: i,j
    do j=1,size(a,2)
       do i=1,size(a,1)
          abs_vvdst(i,j) = abs_dst(a(i,j))
       enddo
    enddo
  end function abs_vvdst
  !
  type (single_st) function max_st_st(a,b)
    implicit none
    type (single_st), intent(in) :: a,b
    type (single_st) :: aux
    logical r
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       if (a%acc.eq.-1) call accuracy_st(a) 
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (a%acc.ge.b%acc) then
          max_st_st = a
       else
          max_st_st = b
       endif
    else
       if ( (a%x+a%y+a%z).ge.(b%x+b%y+b%z) ) then
          max_st_st = a
       else 
          max_st_st = b
       endif
    endif
  end function max_st_st
  !
  type (single_st) function max_st_single(a,b)
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    type (single_st) :: aux
    logical r
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       max_st_single%x = b
       max_st_single%y = b
       max_st_single%z = b
       max_st_single%acc = 7
    else
       if ( (a%x+a%y+a%z).ge.(3.0*b) ) then
          max_st_single = a
       else 
          max_st_single%x = b
          max_st_single%y = b
          max_st_single%z = b
          max_st_single%acc = 7
       endif
    endif
  end function max_st_single
  !
  type (single_st) function max_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    type (single_st) :: aux
    logical r
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       max_single_st%x = a
       max_single_st%y = a
       max_single_st%z = a
       max_single_st%acc = 7
    else
       if ( (3.0*a).ge.(b%x+b%y+b%z) ) then
          max_single_st%x = a
          max_single_st%y = a
          max_single_st%z = a
          max_single_st%acc = 7
       else 
          max_single_st = b
       endif
    endif
  end function max_single_st
  !
  type (double_st) function max_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a,b
    type (double_st) :: aux
    logical r
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       if (a%acc.eq.-1) call accuracy_dst(a) 
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (a%acc.ge.b%acc) then
          max_dst_dst = a
       else
          max_dst_dst = b
       endif
    else
       if ( (a%x+a%y+a%z).ge.(b%x+b%y+b%z) ) then
          max_dst_dst = a
       else 
          max_dst_dst = b
       endif
    endif
  end function max_dst_dst
  !
  type (double_st) function max_dst_double(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    logical r
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       max_dst_double%x = b
       max_dst_double%y = b
       max_dst_double%z = b
       max_dst_double%acc = 7
    else
       if ( (a%x+a%y+a%z).ge.(3.0*b) ) then
          max_dst_double = a
       else 
          max_dst_double%x = b
          max_dst_double%y = b
          max_dst_double%z = b
          max_dst_double%acc = 7
       endif
    endif
  end function max_dst_double
  !
  type (double_st) function max_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    logical r
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       max_double_dst%x = a
       max_double_dst%y = a
       max_double_dst%z = a
       max_double_dst%acc = 7
    else
       if ( (3.0*a).ge.(b%x+b%y+b%z) ) then
          max_double_dst%x = a
          max_double_dst%y = a
          max_double_dst%z = a
          max_double_dst%acc = 7
       else 
          max_double_dst = b
       endif
    endif
  end function max_double_dst
  !
  type (double_st) function max_dst_dst_dst(a,b,c)
    implicit none
    type (double_st), intent(in) :: a,b,c
    max_dst_dst_dst=max_dst_dst(a,max_dst_dst(b,c))
  end function max_dst_dst_dst
  !
  type (double_st) function max_dst_dst_dst_dst(a,b,c,d)
    implicit none
    type (double_st), intent(in) :: a,b,c,d
    max_dst_dst_dst_dst=max_dst_dst(a,max_dst_dst_dst(b,c,d))
  end function max_dst_dst_dst_dst
  !
  type (double_st) function max_dst_dst_dst_dst_dst(a,b,c,d,e)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e
    max_dst_dst_dst_dst_dst=max_dst_dst(a,max_dst_dst_dst_dst(b,c,d,e))
  end function max_dst_dst_dst_dst_dst
  !
  type (double_st) function max_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f
    max_dst_dst_dst_dst_dst_dst=max_dst_dst(a,max_dst_dst_dst_dst_dst(b,c,d,e,f))
  end function max_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function max_dst_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f,g)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f,g
    max_dst_dst_dst_dst_dst_dst_dst=max_dst_dst(a,max_dst_dst_dst_dst_dst_dst(b,c,d,e,f,g))
  end function max_dst_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function max_dst_dst_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f,g,h)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f,g,h
    max_dst_dst_dst_dst_dst_dst_dst_dst=max_dst_dst(a,max_dst_dst_dst_dst_dst_dst_dst(b,c,d,e,f,g,h))
  end function max_dst_dst_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function max_dst_dst_dst_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f,g,h,i)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f,g,h,i
    max_dst_dst_dst_dst_dst_dst_dst_dst_dst=max_dst_dst(a,max_dst_dst_dst_dst_dst_dst_dst_dst(b,c,d,e,f,g,h,i))
  end function max_dst_dst_dst_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function max_dst_dst_double(a,b,c)
    implicit none
    type (double_st), intent(in) :: a,b
    double precision, intent(in) :: c
    max_dst_dst_double=max_dst_dst(a,max_dst_double(b,c))
  end function max_dst_dst_double
  !
  type (double_st) function max_double_dst_dst_dst(a,b,c,d)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b,c,d
    max_double_dst_dst_dst=max_double_dst(a,max_dst_dst_dst(b,c,d))
  end function max_double_dst_dst_dst
  !
  !----------------------------------------------------------------------
  !
  function max_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a,b
    type (double_st), dimension(size(a)) :: max_vdst_vdst
    integer :: i
    do i=1, size(a)
       max_vdst_vdst(i) = max_dst_dst(a(i), b(i))
    enddo
  end function max_vdst_vdst
  !
  function max_vdst_vdouble(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(a)) :: max_vdst_vdouble
    integer :: i
    do i=1, size(a)
       max_vdst_vdouble(i) = max_dst_double(a(i), b(i))
    enddo
  end function max_vdst_vdouble
  !
  function max_vdouble_vdst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: max_vdouble_vdst
    integer :: i
    do i=1, size(b)
       max_vdouble_vdst(i) = max_double_dst(a(i), b(i))
    enddo
  end function max_vdouble_vdst
  !
  function max_vdst_dst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: max_vdst_dst
    integer :: i
    do i=1, size(a)
       max_vdst_dst(i) = max_dst_dst(a(i), b)
    enddo
  end function max_vdst_dst
  !
  function max_dst_vdst(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: max_dst_vdst
    integer :: i
    do i=1, size(b)
       max_dst_vdst(i) = max_dst_dst(a, b(i))
    enddo
  end function max_dst_vdst
  !
  function max_vdst_double(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a)) :: max_vdst_double
    integer :: i
    do i=1, size(a)
       max_vdst_double(i) = max_dst_double(a(i), b)
    enddo
  end function max_vdst_double
  !
  function max_double_vdst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: max_double_vdst
    integer :: i
    do i=1, size(b)
       max_double_vdst(i) = max_double_dst(a, b(i))
    enddo
  end function max_double_vdst
  !
  function max_vdouble_dst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: max_vdouble_dst
    integer :: i
    do i=1, size(a)
       max_vdouble_dst(i) = max_double_dst(a(i), b)
    enddo
  end function max_vdouble_dst
  !
  function max_dst_vdouble(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: max_dst_vdouble
    integer :: i
    do i=1, size(b)
       max_dst_vdouble(i) = max_dst_double(a, b(i))
    enddo
  end function max_dst_vdouble
  !
  type (single_st) function min_st_st(a,b)
    implicit none
    type (single_st), intent(in) :: a,b
    type (single_st) :: aux
    logical r
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = nb_significant_digit_st(aux).eq.0
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       if (a%acc.eq.-1) call accuracy_st(a) 
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (a%acc.ge.b%acc) then
          min_st_st = a
       else
          min_st_st = b
       endif
    else
       if ( (a%x+a%y+a%z).le.(b%x+b%y+b%z) ) then
          min_st_st = a
       else 
          min_st_st = b
       endif
    endif
  end function min_st_st
  !
  type (single_st) function min_st_single(a,b)
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    type (single_st) :: aux
    logical r
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = nb_significant_digit_st(aux).eq.0
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       min_st_single%x = b
       min_st_single%y = b
       min_st_single%z = b
       min_st_single%acc = 7
    else
       if ( (a%x+a%y+a%z).le.(3.0*b) ) then
          min_st_single = a
       else 
          min_st_single%x = b
          min_st_single%y = b
          min_st_single%z = b
          min_st_single%acc = 7
       endif
    endif
  end function min_st_single
  !
  type (single_st) function min_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    type (single_st) :: aux
    logical r
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = nb_significant_digit_st(aux).eq.0
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       min_single_st%x = a
       min_single_st%y = a
       min_single_st%z = a
       min_single_st%acc = 7
    else
       if ( (3.0*a).le.(b%x+b%y+b%z) ) then
          min_single_st%x = a
          min_single_st%y = a
          min_single_st%z = a
          min_single_st%acc = 7
       else 
          min_single_st = b
       endif
    endif
  end function min_single_st
  !
  type (double_st) function min_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a,b
    type (double_st) :: aux
    logical r
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = nb_significant_digit_dst(aux).eq.0
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       if (a%acc.eq.-1) call accuracy_dst(a) 
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (a%acc.ge.b%acc) then
          min_dst_dst = a
       else
          min_dst_dst = b
       endif
    else
       if ( (a%x+a%y+a%z).le.(b%x+b%y+b%z) ) then
          min_dst_dst = a
       else 
          min_dst_dst = b
       endif
    endif
  end function min_dst_dst
  !
  type (double_st) function min_dst_double(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    logical r
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = nb_significant_digit_dst(aux).eq.0
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       min_dst_double%x = b
       min_dst_double%y = b
       min_dst_double%z = b
       min_dst_double%acc = 7
    else
       if ( (a%x+a%y+a%z).le.(3.0*b) ) then
          min_dst_double = a
       else 
          min_dst_double%x = b
          min_dst_double%y = b
          min_dst_double%z = b
          min_dst_double%acc = 7
       endif
    endif
  end function min_dst_double
  !
  type (double_st) function min_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    logical r
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    if (r) then
       min_double_dst%x = a
       min_double_dst%y = a
       min_double_dst%z = a
       min_double_dst%acc = 7
    else
       if ( (3.0*a).le.(b%x+b%y+b%z) ) then
          min_double_dst%x = a
          min_double_dst%y = a
          min_double_dst%z = a
          min_double_dst%acc = 7
       else 
          min_double_dst = b
       endif
    endif
  end function min_double_dst
  !
  type (double_st) function min_dst_dst_dst(a,b,c)
    implicit none
    type (double_st), intent(in) :: a,b,c
    min_dst_dst_dst=min_dst_dst(a,min_dst_dst(b,c))
  end function min_dst_dst_dst
  !
  type (double_st) function min_dst_dst_dst_dst(a,b,c,d)
    implicit none
    type (double_st), intent(in) :: a,b,c,d
    min_dst_dst_dst_dst=min_dst_dst(a,min_dst_dst_dst(b,c,d))
  end function min_dst_dst_dst_dst
  !
  type (double_st) function min_dst_dst_dst_dst_dst(a,b,c,d,e)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e
    min_dst_dst_dst_dst_dst=min_dst_dst(a,min_dst_dst_dst_dst(b,c,d,e))
  end function min_dst_dst_dst_dst_dst
  !
  type (double_st) function min_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f
    min_dst_dst_dst_dst_dst_dst=min_dst_dst(a,min_dst_dst_dst_dst_dst(b,c,d,e,f))
  end function min_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function min_dst_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f,g)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f,g
    min_dst_dst_dst_dst_dst_dst_dst=min_dst_dst(a,min_dst_dst_dst_dst_dst_dst(b,c,d,e,f,g))
  end function min_dst_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function min_dst_dst_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f,g,h)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f,g,h
    min_dst_dst_dst_dst_dst_dst_dst_dst=min_dst_dst(a,min_dst_dst_dst_dst_dst_dst_dst(b,c,d,e,f,g,h))
  end function min_dst_dst_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function min_dst_dst_dst_dst_dst_dst_dst_dst_dst(a,b,c,d,e,f,g,h,i)
    implicit none
    type (double_st), intent(in) :: a,b,c,d,e,f,g,h,i
    min_dst_dst_dst_dst_dst_dst_dst_dst_dst=min_dst_dst(a,min_dst_dst_dst_dst_dst_dst_dst_dst(b,c,d,e,f,g,h,i))
  end function min_dst_dst_dst_dst_dst_dst_dst_dst_dst
  !
  type (double_st) function min_dst_dst_double(a,b,c)
    implicit none
    type (double_st), intent(in) :: a,b
    double precision, intent(in) :: c
    min_dst_dst_double=min_dst_dst(a,min_dst_double(b,c))
  end function min_dst_dst_double
  !
!----------------------------------------------------------------------
!
  function min_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a,b
    type (double_st), dimension(size(a)) :: min_vdst_vdst
    integer :: i
    do i=1, size(a)
       min_vdst_vdst(i) = min_dst_dst(a(i), b(i))
    enddo
  end function min_vdst_vdst
  !
  function min_vdst_vdouble(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(a)) :: min_vdst_vdouble
    integer :: i
    do i=1, size(a)
       min_vdst_vdouble(i) = min_dst_double(a(i), b(i))
    enddo
  end function min_vdst_vdouble
  !
  function min_vdouble_vdst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: min_vdouble_vdst
    integer :: i
    do i=1, size(b)
       min_vdouble_vdst(i) = min_double_dst(a(i), b(i))
    enddo
  end function min_vdouble_vdst
  !
  function min_vdst_dst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: min_vdst_dst
    integer :: i
    do i=1, size(a)
       min_vdst_dst(i) = min_dst_dst(a(i), b)
    enddo
  end function min_vdst_dst
  !
  function min_dst_vdst(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: min_dst_vdst
    integer :: i
    do i=1, size(b)
       min_dst_vdst(i) = min_dst_dst(a, b(i))
    enddo
  end function min_dst_vdst
  !
  function min_vdst_double(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a)) :: min_vdst_double
    integer :: i
    do i=1, size(a)
       min_vdst_double(i) = min_dst_double(a(i), b)
    enddo
  end function min_vdst_double
  !
  function min_double_vdst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: min_double_vdst
    integer :: i
    do i=1, size(b)
       min_double_vdst(i) = min_double_dst(a, b(i))
    enddo
  end function min_double_vdst
  !
  function min_vdouble_dst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: min_vdouble_dst
    integer :: i
    do i=1, size(a)
       min_vdouble_dst(i) = min_double_dst(a(i), b)
    enddo
  end function min_vdouble_dst
  !
  function min_dst_vdouble(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: min_dst_vdouble
    integer :: i
    do i=1, size(b)
       min_dst_vdouble(i) = min_dst_double(a, b(i))
    enddo
  end function min_dst_vdouble
 !
 type (single_st) function sign_st_st(a,b)
   implicit none
   type (single_st), intent(in) :: a,b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_st(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   sign_st_st%x = sign(a%x ,b%x) 
   sign_st_st%y = sign(a%y ,b%y) 
   sign_st_st%z = sign(a%z ,b%z)
   sign_st_st%acc = -1 
 end function sign_st_st
 !
 type (single_st) function sign_single_st(a,b)
   implicit none
   real, intent(in) :: a
   type (single_st), intent(in) :: b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_st(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   sign_single_st%x = sign(a ,b%x) 
   sign_single_st%y = sign(a ,b%y) 
   sign_single_st%z = sign(a ,b%z) 
   sign_single_st%acc = -1
 end function sign_single_st
 !
 type (single_st) function sign_st_single(a,b)
   implicit none
   type (single_st), intent(in) :: a
   real, intent(in) :: b
   sign_st_single%x = sign(a%x ,b) 
   sign_st_single%y = sign(a%y ,b) 
   sign_st_single%z = sign(a%z ,b) 
   sign_st_single%acc = -1
 end function sign_st_single
 !
 type (double_st) function sign_dst_dst(a,b)
   implicit none
   type (double_st), intent(in) :: a,b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_dst(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   sign_dst_dst%x = sign(a%x ,b%x) 
   sign_dst_dst%y = sign(a%y ,b%y) 
   sign_dst_dst%z = sign(a%z ,b%z)
   sign_dst_dst%acc = -1 
 end function sign_dst_dst
 !
 type (double_st) function sign_double_dst(a,b)
   implicit none
   double precision, intent(in) :: a
   type (double_st), intent(in) :: b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_dst(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   sign_double_dst%x = sign(a ,b%x) 
   sign_double_dst%y = sign(a ,b%y) 
   sign_double_dst%z = sign(a ,b%z) 
   sign_double_dst%acc = -1
 end function sign_double_dst
 !
 type (double_st) function sign_dst_double(a,b)
   implicit none
   type (double_st), intent(in) :: a
   double precision, intent(in) :: b
   sign_dst_double%x = sign(a%x ,b) 
   sign_dst_double%y = sign(a%y ,b) 
   sign_dst_double%z = sign(a%z ,b) 
   sign_dst_double%acc = -1 
 end function sign_dst_double
 !
 type (single_st) function mod_st_st(a,b)
   implicit none
   type (single_st), intent(in) :: a,b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_st(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   mod_st_st%x = mod(a%x ,b%x) 
   mod_st_st%y = mod(a%y ,b%y) 
   mod_st_st%z = mod(a%z ,b%z)
   mod_st_st%acc = -1 
 end function mod_st_st
 !
 type (single_st) function mod_single_st(a,b)
   implicit none
   real, intent(in) :: a
   type (single_st), intent(in) :: b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_st(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   mod_single_st%x = mod(a ,b%x) 
   mod_single_st%y = mod(a ,b%y) 
   mod_single_st%z = mod(a ,b%z) 
   mod_single_st%acc = -1
 end function mod_single_st
 !
 type (single_st) function mod_st_single(a,b)
   implicit none
   type (single_st), intent(in) :: a
   real, intent(in) :: b
   mod_st_single%x = mod(a%x ,b) 
   mod_st_single%y = mod(a%y ,b) 
   mod_st_single%z = mod(a%z ,b) 
   mod_st_single%acc = -1
 end function mod_st_single
 !
 type (double_st) function mod_dst_dst(a,b)
   implicit none
   type (double_st), intent(in) :: a,b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_dst(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   mod_dst_dst%x = mod(a%x ,b%x) 
   mod_dst_dst%y = mod(a%y ,b%y) 
   mod_dst_dst%z = mod(a%z ,b%z)
   mod_dst_dst%acc = -1 
 end function mod_dst_dst
 !
 type (double_st) function mod_double_dst(a,b)
   implicit none
   double precision, intent(in) :: a
   type (double_st), intent(in) :: b
   if (cadna_intrinsic_tag) then
      if (b%acc.eq.-1) call accuracy_dst(b) 
      if (b%acc.eq.0) then
         call instability(cadna_intrinsic)
      end if
   end if
   mod_double_dst%x = mod(a ,b%x) 
   mod_double_dst%y = mod(a ,b%y) 
   mod_double_dst%z = mod(a ,b%z) 
   mod_double_dst%acc = -1
 end function mod_double_dst
 !
 type (double_st) function mod_dst_double(a,b)
   implicit none
   type (double_st), intent(in) :: a
   double precision, intent(in) :: b
   mod_dst_double%x = mod(a%x ,b) 
   mod_dst_double%y = mod(a%y ,b) 
   mod_dst_double%z = mod(a%z ,b) 
   mod_dst_double%acc = -1 
 end function mod_dst_double
 
 type (double_st) function modulo_dst_double(a,b)
   implicit none
   type (double_st), intent(in) :: a
   double precision, intent(in) :: b
   modulo_dst_double%x = modulo(a%x ,b) 
   modulo_dst_double%y = modulo(a%y ,b) 
   modulo_dst_double%z = modulo(a%z ,b) 
   modulo_dst_double%acc = -1 
 end function modulo_dst_double
 
 !
 type (single_st) function dim_st_st(a,b)
   implicit none
   type (single_st), intent(in) :: a,b
   type (single_st) :: aux
   logical r
   if (cadna_intrinsic_tag) then
      aux%x = a%x - b%x
      aux%y = a%y - b%y
      aux%z = a%z - b%z
      r = computed_zero_st(aux)
      if (r) then
         call instability(cadna_intrinsic)
      endif
   endif
   dim_st_st%x = dim(a%x ,b%x) 
   dim_st_st%y = dim(a%y ,b%y) 
   dim_st_st%z = dim(a%z ,b%z) 
   dim_st_st%acc = -1 
 end function dim_st_st
 !
 type (single_st) function dim_single_st(a,b)
   implicit none
   real, intent(in) :: a
   type (single_st), intent(in) :: b
   type (single_st) :: aux
   logical r
   if (cadna_intrinsic_tag) then
      aux%x = a - b%x
      aux%y = a - b%y
      aux%z = a - b%z
      r = computed_zero_st(aux)
      if (r) then
         call instability(cadna_intrinsic)
      endif
   endif
   dim_single_st%x = dim(a ,b%x) 
   dim_single_st%y = dim(a ,b%y) 
   dim_single_st%z = dim(a ,b%z) 
   dim_single_st%acc = -1
 end function dim_single_st
 !
 type (single_st) function dim_st_single(a,b)
   implicit none
   type (single_st), intent(in) :: a
   real, intent(in) :: b
   type (single_st) :: aux
   logical r
   if (cadna_intrinsic_tag) then
      aux%x = a%x - b
      aux%y = a%y - b
      aux%z = a%z - b
      r = computed_zero_st(aux)
      if (r) then
         call instability(cadna_intrinsic)
      endif
   endif
   dim_st_single%x = dim(a%x ,b) 
   dim_st_single%y = dim(a%y ,b) 
   dim_st_single%z = dim(a%z ,b) 
   dim_st_single%acc = -1
 end function dim_st_single
 !
 type (double_st) function dim_dst_dst(a,b)
   implicit none
   type (double_st), intent(in) :: a,b
   type (double_st) :: aux
   logical r
   if (cadna_intrinsic_tag) then
      aux%x = a%x - b%x
      aux%y = a%y - b%y
      aux%z = a%z - b%z
      r = computed_zero_dst(aux)
      if (r) then
         call instability(cadna_intrinsic)
      endif
   endif
   dim_dst_dst%x = dim(a%x ,b%x) 
   dim_dst_dst%y = dim(a%y ,b%y) 
   dim_dst_dst%z = dim(a%z ,b%z) 
   dim_dst_dst%acc = -1 
 end function dim_dst_dst
 !
 type (double_st) function dim_double_dst(a,b)
   implicit none
   double precision, intent(in) :: a
   type (double_st), intent(in) :: b
   type (double_st) :: aux
   logical r
   if (cadna_intrinsic_tag) then
      aux%x = a - b%x
      aux%y = a - b%y
      aux%z = a - b%z
      r = computed_zero_dst(aux)
      if (r) then
         call instability(cadna_intrinsic)
      endif
   endif
   dim_double_dst%x = dim(a ,b%x) 
   dim_double_dst%y = dim(a ,b%y) 
   dim_double_dst%z = dim(a ,b%z) 
   dim_double_dst%acc = -1
 end function dim_double_dst
 !
 type (double_st) function dim_dst_double(a,b)
   implicit none
   type (double_st), intent(in) :: a
   double precision, intent(in) :: b
   type (double_st) :: aux
   logical r
   if (cadna_intrinsic_tag) then
      aux%x = a%x - b
      aux%y = a%y - b
      aux%z = a%z - b
      r = computed_zero_dst(aux)
      if (r) then
         call instability(cadna_intrinsic)
      endif
   endif
   dim_dst_double%x = dim(a%x ,b) 
   dim_dst_double%y = dim(a%y ,b) 
   dim_dst_double%z = dim(a%z ,b) 
   dim_dst_double%acc = -1
 end function dim_dst_double
 !
 type (double_st) function minval_vdst(a, mask)
   use cadna_lt
   implicit none
   type (double_st), dimension(:), intent(in) :: a
   logical, dimension(:), intent(in), optional :: mask
   integer :: i, i0
   if (present(mask)) then
      minval_vdst=a(1)
      do i=2,size(a)
         if (a(i)<minval_vdst) then
            minval_vdst=a(i)
         endif
      enddo
   else
      do i=1,size(a)
         if (mask(i)) then
            i0=i
            minval_vdst=a(i)
            exit
         endif
      enddo
      do i=i0+1,size(a)
         if (mask(i)) then
            if (a(i)<minval_vdst) then
               minval_vdst=a(i)
            endif
         endif
      enddo
   endif
 end function minval_vdst
 !
 function minval_vvdst_dim(a, dim)
   implicit none
   type (double_st), dimension(:,:), intent(in) :: a
   integer, intent(in) :: dim
   type (double_st), dimension(size(a,3-dim)) :: minval_vvdst_dim
   integer :: i   
   if (dim==1) then      
      do i=1,size(a,2)
         minval_vvdst_dim(i)=minval_vdst(a(:,i))
      enddo
   else
      do i=1,size(a,1)
         minval_vvdst_dim(i)=minval_vdst(a(i,:))
      enddo
   endif
 end function minval_vvdst_dim
 !
 function minval_vvdst(a)
   use cadna_lt
   implicit none
   type (double_st), dimension(:,:), intent(in) :: a
   type (double_st) :: minval_vvdst
   integer :: i,j
   minval_vvdst=a(1,1)
   do j=1,size(a,2)
      do i=1,size(a,1)
         if (a(i,j)<minval_vvdst) then
            minval_vvdst=a(i,j)
         endif
      enddo
   enddo
 end function minval_vvdst
 !
 type (double_st) function maxval_vdst(a)
   use cadna_gt
   implicit none
   type (double_st), dimension(:), intent(in) :: a
   integer :: i
   maxval_vdst=a(1)
   do i=2,size(a)
      if (a(i)>maxval_vdst) then
         maxval_vdst=a(i)
      endif
   enddo
 end function maxval_vdst
 !
 function maxval_vvdst_dim(a, dim)
   implicit none
   type (double_st), dimension(:,:), intent(in) :: a
   integer, intent(in) :: dim
   type (double_st), dimension(size(a,3-dim)) :: maxval_vvdst_dim
   integer :: i   
   if (dim==1) then      
      do i=1,size(a,2)
         maxval_vvdst_dim(i)=maxval_vdst(a(:,i))
      enddo
   else
      do i=1,size(a,1)
         maxval_vvdst_dim(i)=maxval_vdst(a(i,:))
      enddo
   endif
 end function maxval_vvdst_dim
 !
 function maxval_vvdst(a)
   use cadna_gt
   implicit none
   type (double_st), dimension(:,:), intent(in) :: a
   type (double_st) :: maxval_vvdst
   integer :: i,j
   maxval_vvdst=a(1,1)
   do j=1,size(a,2)
      do i=1,size(a,1)
         if (a(i,j)>maxval_vvdst) then
            maxval_vvdst=a(i,j)
         endif
      enddo
   enddo
 end function maxval_vvdst
 !
 function matmul_vvdst_vdst(a,b)
   use cadna_mul
   use cadna_add
   implicit none
   type (double_st), dimension(:,:), intent(in) :: a
   type (double_st), dimension(:), intent(in) :: b
   type (double_st), dimension(size(a,1)) :: matmul_vvdst_vdst
   integer :: i, j, n, m

   n=size(a,1)
   m=size(a,2)
   matmul_vvdst_vdst=double_st(0.D0, 0.D0, 0.D0, -1, 0)
!   matmul_vvdst_vdst=0.D0
   do j=1,m
      do i=1,n
         matmul_vvdst_vdst(i)%x=matmul_vvdst_vdst(i)%x+a(i,j)%x*b(j)%x
         matmul_vvdst_vdst(i)%y=matmul_vvdst_vdst(i)%y+a(i,j)%x*b(j)%y
         matmul_vvdst_vdst(i)%z=matmul_vvdst_vdst(i)%z+a(i,j)%x*b(j)%z
      enddo
   enddo

 end function matmul_vvdst_vdst
 !
 function matmul_vdst_vvdst(a,b)
   use cadna_mul
   use cadna_add
   implicit none
   type (double_st), dimension(:), intent(in) :: a
   type (double_st), dimension(:,:), intent(in) :: b
   type (double_st), dimension(size(b,2)) :: matmul_vdst_vvdst

   integer :: i, j, n, m

   n=size(b,1)
   m=size(b,2)
   matmul_vdst_vvdst=double_st(0.D0, 0.D0, 0.D0, -1, 0)
!   matmul_vdst_vvdst=0.D0
   do j=1,m
      do i=1,n
!         matmul_vdst_vvdst(i)=matmul_vdst_vvdst(i)+a(i)*b(i,j)
         matmul_vdst_vvdst(i)%x=matmul_vdst_vvdst(i)%x+a(i)%x*b(i,j)%x
         matmul_vdst_vvdst(i)%y=matmul_vdst_vvdst(i)%y+a(i)%y*b(i,j)%y
         matmul_vdst_vvdst(i)%z=matmul_vdst_vvdst(i)%z+a(i)%z*b(i,j)%z
      enddo
   enddo

 end function matmul_vdst_vvdst
 !
 function matmul_vvdst_vvdst(a,b)
   use cadna_mul
   use cadna_add
   implicit none
   type (double_st), dimension(:,:), intent(in) :: a
   type (double_st), dimension(:,:), intent(in) :: b
   type (double_st), dimension(size(a,1), size(b,2)) :: matmul_vvdst_vvdst
   integer :: i, j, k, n, m, p

   n=size(a,1)
   m=size(a,2)
   p=size(b,2)
   matmul_vvdst_vvdst=double_st(0.D0, 0.D0, 0.D0, -1, 0)
   do k=1,p
      do j=1,m
         do i=1,n
!            matmul_vvdst_vvdst(i,k)=matmul_vvdst_vvdst(i,k)+a(i,j)*b(j,k)
            matmul_vvdst_vvdst(i,k)%x=matmul_vvdst_vvdst(i,k)%x+a(i,j)%x*b(j,k)%x
            matmul_vvdst_vvdst(i,k)%y=matmul_vvdst_vvdst(i,k)%y+a(i,j)%y*b(j,k)%y
            matmul_vvdst_vvdst(i,k)%z=matmul_vvdst_vvdst(i,k)%z+a(i,j)%z*b(j,k)%z
         enddo
      enddo
   enddo

 end function matmul_vvdst_vvdst
 !
 function dot_product_vdst_vdst(a,b)
   use cadna_mul
   use cadna_add
   implicit none
   type (double_st), dimension(:), intent(in) :: a
   type (double_st), dimension(:), intent(in) :: b
   type (double_st) :: dot_product_vdst_vdst
   integer :: i, n
   n=size(a)   
   dot_product_vdst_vdst=double_st(0.D0, 0.D0, 0.D0, -1, 0)
!   dot_product_vdst_vdst=0.D0
   do i=1,n
!      dot_product_vdst_vdst=dot_product_vdst_vdst+a(i)*b(i)
      dot_product_vdst_vdst%x=dot_product_vdst_vdst%x+a(i)%x*b(i)%x
      dot_product_vdst_vdst%y=dot_product_vdst_vdst%y+a(i)%y*b(i)%y
      dot_product_vdst_vdst%z=dot_product_vdst_vdst%z+a(i)%z*b(i)%z
   enddo

 end function dot_product_vdst_vdst
 !
 function huge_dst(a)
   implicit none
   type (double_st), intent(in) :: a
   double precision :: huge_dst
   huge_dst=huge(a%x)
 end function huge_dst
 !
 function sum_dst_vdst(a)
   use cadna_add
   implicit none
   type (double_st), dimension(:), intent(in) :: a
   type (double_st) :: sum_dst_vdst
   integer :: i, n
   sum_dst_vdst=double_st(0.D0, 0.D0, 0.D0, -1, 0)
!   sum_dst_vdst=0.D0
   n=size(a)
   do i=1,n
!      sum_dst_vdst=sum_dst_vdst+a(i)
      sum_dst_vdst%x=sum_dst_vdst%x+a(i)%x
      sum_dst_vdst%y=sum_dst_vdst%y+a(i)%y
      sum_dst_vdst%z=sum_dst_vdst%z+a(i)%z
   enddo
 end function sum_dst_vdst
 !
 integer function floor_dst(a)
   use cadna_to
   implicit none
   type (double_st), intent(in) :: a
   double precision :: b
   b=a
   floor_dst=floor(b)
  end function floor_dst
 !
 integer function ceiling_dst(a)
   use cadna_to
   implicit none
   type (double_st), intent(in) :: a
   double precision :: b
   b=a
   ceiling_dst=ceiling(b)
  end function ceiling_dst
  !
  subroutine cpu_time_dst(a)
    use cadna_to
    implicit none
    type (double_st), intent(out) :: a
    double precision :: b
    call cpu_time(b)
    a=b
  end subroutine cpu_time_dst
  !
  double precision function epsilon_dst(a)
    use cadna_to
    implicit none
    type (double_st), intent(in) :: a
    double precision :: b
    b=a
    epsilon_dst=epsilon(b)
  end function epsilon_dst
  !
  function nint_vdst(a)
    use cadna_to
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    integer, dimension(size(a)) :: nint_vdst
    integer :: i
    double precision :: b
    do i=1,size(a)
       b=a(i)
       nint_vdst(i)=nint(b)
    enddo
  end function nint_vdst
  !
  function nint_vvdst(a)
    use cadna_to
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a
    integer, dimension(size(a,1),size(a,2)) :: nint_vvdst
    integer :: i,j
    double precision :: b
    do j=1,size(a,2)
       do i=1,size(a,1)
          b=a(i,j)
          nint_vvdst(i,j)=nint(b)
       enddo
    enddo
  end function nint_vvdst
  !
  integer function maxloc_vdst(a)
    use cadna_gt
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    integer :: i
    maxloc_vdst=1
    do i=2,size(a)
       if (a(i)>a(maxloc_vdst)) then
          maxloc_vdst=i
       endif
    enddo
  end function maxloc_vdst
  !
  function maxloc_vvdst(a)
    use cadna_gt
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a
    integer, dimension(2) :: maxloc_vvdst
    integer :: i,j
    maxloc_vvdst(1)=1
    maxloc_vvdst(2)=1
    do j=1,size(a,2)
       do i=1,size(a,1)
          if (a(i,j)>a(maxloc_vvdst(1),maxloc_vvdst(2))) then
             maxloc_vvdst(1)=i
             maxloc_vvdst(2)=j
          endif
       enddo
    enddo
  end function maxloc_vvdst
  !
  integer function minloc_vdst(a)
    use cadna_lt
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    integer :: i
    minloc_vdst=1
    do i=2,size(a)
       if (a(i)<a(minloc_vdst)) then
          minloc_vdst=i
       endif
    enddo
  end function minloc_vdst
  !
  function minloc_vvdst(a)
    use cadna_lt
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a
    integer, dimension(2) :: minloc_vvdst
    integer :: i,j
    minloc_vvdst(1)=1
    minloc_vvdst(2)=1
    do j=1,size(a,2)
       do i=1,size(a,1)
          if (a(i,j)<a(minloc_vvdst(1),minloc_vvdst(2))) then
             minloc_vvdst(1)=i
             minloc_vvdst(2)=j
          endif
       enddo
    enddo
  end function minloc_vvdst
  !
  function sign_double_vvdst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), dimension(:,:), intent(in) :: b
    type (double_st), dimension(size(b,1),size(b,2)) :: sign_double_vvdst
    integer :: i,j
    do j=1,size(b,2)
       do i=1,size(b,1)
          sign_double_vvdst(i,j)=sign_double_dst(a,b(i,j))
       enddo
    enddo
  end function sign_double_vvdst
 !
  function max_vvvdst_double(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: max_vvvdst_double
    integer :: i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             max_vvvdst_double(i,j,k)=max_dst_double(a(i,j,k),b)
          enddo
       enddo
    enddo
  end function max_vvvdst_double
 !
  function min_vvvdst_double(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: min_vvvdst_double
    integer :: i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             min_vvvdst_double(i,j,k)=min_dst_double(a(i,j,k),b)
          enddo
       enddo
    enddo
  end function min_vvvdst_double
  !
  function abs_vvvdst(a)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: abs_vvvdst
    integer :: i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             abs_vvvdst(i,j,k) = abs_dst(a(i,j,k))
          enddo
       enddo
    enddo
  end function abs_vvvdst
  !
  function max_double_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b,1),size(b,2),size(b,3)) :: max_double_vvvdst
    integer :: i,j,k
    max_double_vvvdst=max_vvvdst_double(b,a)
  end function max_double_vvvdst
 !
  function min_double_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b,1),size(b,2),size(b,3)) :: min_double_vvvdst
    integer :: i,j,k
    min_double_vvvdst=min_vvvdst_double(b,a)
  end function min_double_vvvdst
 !
  function max_vvvdst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a, b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: max_vvvdst_vvvdst
    integer :: i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             max_vvvdst_vvvdst(i,j,k)=max_dst_dst(a(i,j,k),b(i,j,k))
          enddo
       enddo
    enddo
  end function max_vvvdst_vvvdst
 !
  function min_vvvdst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a, b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: min_vvvdst_vvvdst
    integer :: i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             min_vvvdst_vvvdst(i,j,k)=min_dst_dst(a(i,j,k),b(i,j,k))
          enddo
       enddo
    enddo
  end function min_vvvdst_vvvdst
 !
 function maxval_vvvdst(a)
   use cadna_gt
   implicit none
   type (double_st), dimension(:,:,:), intent(in) :: a
   type (double_st) :: maxval_vvvdst
   integer :: i,j,k
   maxval_vvvdst=a(1,1,1)
   do k=1,size(a,3)
      do j=1,size(a,2)
         do i=1,size(a,1)
            if (a(i,j,k)>maxval_vvvdst) then
               maxval_vvvdst=a(i,j,k)
            endif
         enddo
      enddo
   enddo
 end function maxval_vvvdst
 !
end module cadna_intr
