!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_le
!  NAME
!    cadna_le
!
!  DESCRIPTION
!    Define all the functions involving at least one argument
!    of stochastic type which overload the "<=" operator
!    in a test such as "a<=b".
!
!  INPUTS
!    All the functions have two arguments of stochastic, 
!    real or integer type in single or double precision. At least 
!    one argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one logical return value. 
!
!  NOTES
!    The return value takes into account the stochastic definition of
!    the "<=" operator (see the documentation).
!
!******
!
module cadna_le
  !
  use cadna_type
  use cadna_cestac
  !
  interface operator (<=)
     !	=======================
     module procedure le_st_st, &
          &		le_single_st, le_st_single, le_int_st, le_st_int, &
          &		le_dst_dst, le_dst_double, le_double_dst, &
          &		le_dst_st, le_st_dst, le_dst_single, le_single_dst, &
          &		le_st_double, le_double_st, le_int_dst, le_dst_int
  end interface
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  logical function le_st_st(a,b)
    type  (single_st), intent(in) :: a, b
    logical :: r
    type  (single_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_st_st = (r.or.( (a%x+a%y+a%z).le.(b%x+b%y+b%z) ))
  end function le_st_st
  !
  logical function le_single_st(a,b)
    type  (single_st), intent(in) :: b
    real, intent(in) :: a
    logical r
    type  (single_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_single_st = (r.or.( (a+a+a).le.(b%x+b%y+b%z) ))
  end function le_single_st
  !
  logical function le_st_single(a,b)
    type  (single_st), intent(in) :: a
    real, intent(in) :: b
    logical r
    type  (single_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_st_single = (r.or.( (a%x+a%y+a%z).le.(b+b+b) ))
  end function le_st_single
  !
  logical function le_int_st(a,b)
    type  (single_st), intent(in) :: b
    integer, intent(in) :: a
    logical r
    type  (single_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_int_st = (r.or.( (a+a+a).le.(b%x+b%y+b%z) ))
  end function le_int_st
  !
  logical function le_st_int(a,b)
    type  (single_st), intent(in) :: a
    integer, intent(in) :: b
    logical r
    type  (single_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_st(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_st_int = (r.or.( (a%x+a%y+a%z).le.(b+b+b) ))
  end function le_st_int
  !
  logical function le_dst_dst(a,b)
    type (double_st), intent(in) :: a,b
    logical :: r
    type  (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_dst_dst = (r.or.( (a%x+a%y+a%z).le.(b%x+b%y+b%z) ))
  end function le_dst_dst
  !
  logical function le_dst_double(a,b)
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_dst_double = (r.or.( (a%x+a%y+a%z).le.(b+b+b) ))
  end function le_dst_double
  !
  logical function le_double_dst(a,b)
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_double_dst = (r.or.( (a+a+a).le.(b%x+b%y+b%z) ))
  end function le_double_dst
  !
  logical function le_dst_st(a,b)
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_dst_st = (r.or.( (a%x+a%y+a%z).le.(b%x+b%y+b%z) ))
  end function le_dst_st
  !
  logical function le_st_dst(a,b)
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a%x - b%x
    aux%y = a%y - b%y
    aux%z = a%z - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_st_dst = (r.or.( (a%x+a%y+a%z).le.(b%x+b%y+b%z) ))
  end function le_st_dst
  !
  logical function le_single_dst(a,b)
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_single_dst = (r.or.( (a+a+a).le.(b%x+b%y+b%z) ))
  end function le_single_dst
  !
  logical function le_dst_single(a,b)
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_dst_single = (r.or.( (a%x+a%y+a%z).le.(b+b+b) ))
  end function le_dst_single
  !
  logical function le_st_double(a,b)
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_st_double = (r.or.( (a%x+a%y+a%z).le.(b+b+b) ))
  end function le_st_double
  !
  logical function le_double_st(a,b)
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_double_st = (r.or.( (a+a+a).le.(b%x+b%y+b%z) ))
  end function le_double_st
  !
  logical function le_int_dst(a,b)
    type  (double_st), intent(in) :: b
    integer, intent(in) :: a
    logical r
    type  (double_st) :: aux
    aux%x = a - b%x
    aux%y = a - b%y
    aux%z = a - b%z
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
   le_int_dst = (r.or.( (a+a+a).le.(b%x+b%y+b%z) ))
  end function le_int_dst
  !
  logical function le_dst_int(a,b)
    type  (double_st), intent(in) :: a
    integer, intent(in) :: b
    logical r
    type  (double_st) :: aux
    aux%x = a%x - b
    aux%y = a%y - b
    aux%z = a%z - b
    r = computed_zero_dst(aux)
    if (cadna_branching_tag) then
       if (r) then
          call instability(cadna_branching)
       endif
    endif
    le_dst_int = (r.or.( (a%x+a%y+a%z).le.(b+b+b) ))
  end function le_dst_int
  !
end module cadna_le
