!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_sub
!  NAME
!    cadna_sub
!
!
!  DESCRIPTION
!    Defines all the functions involving at least one argument
!    of stochastic type which overload the "-" operator
!    in a statement such as "a-b" or "-a".
!
!  INPUTS
!    All the functions have one or two arguments of stochastic, 
!    real or integer type in single or double precision. At least 
!    one argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one numerical output of stochastic type. 
!    It can be an array of rank 1 if one input argument is an array 
!    of rank 1. The output is in double precision if, at least,
!    one input is in double precision.
!
!  NOTES
!    If one argumment is an array of rank 1, the arguments cannot
!    be of integer type and must be in double precision.

!
!******
!
module cadna_sub
  !
  use cadna_type
  use cadna_cestac
  !
  interface operator (-)
     !       ======================
     module procedure sub_st, sub_dst, sub_st_st, &
          &		sub_single_st, sub_st_single, sub_int_st, &
          &		sub_st_int, sub_dst_dst, sub_dst_double, &
          &		sub_double_dst, sub_dst_st, sub_st_dst, &
          &		sub_dst_single, sub_single_dst, sub_st_double, &
          &		sub_double_st, sub_int_dst, sub_dst_int, &
          !--------------------------------------------------------------
          &          sub_vdst, sub_vdst_vdst, sub_dst_vdst, sub_vdst_dst, &
          &          sub_double_vdst, sub_vdst_double, sub_vdouble_vdst, &
          &          sub_vdst_vdouble, sub_vdouble_dst, sub_dst_vdouble, &
          !-----------------------------------
          &          sub_vvdst_vvdst, sub_vvdst, &
          &          sub_vvvdst_vvvdst, &
          &          sub_vvvdst_double, sub_double_vvvdst
  end interface
  !
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  type (single_st) function sub_st(a)
    implicit none
    type (single_st), intent(in) :: a
    sub_st%x = - a%x
    sub_st%y = - a%y
    sub_st%z = - a%z
    sub_st%acc = a%acc
  end function sub_st
  !
  type (double_st) function sub_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    sub_dst%x = - a%x
    sub_dst%y = - a%y
    sub_dst%z = - a%z
    sub_dst%acc = a%acc
  end function sub_dst
  !
  type (single_st) function sub_st_st(a,b)
    implicit none
    type (single_st), intent(in) :: a, b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y - b%y
    call rnd_switch
    aux%z = a%z - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_st_st = aux
  end function sub_st_st
  !
  type (single_st) function sub_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a - b%x
    if (myrandom()) call rnd_switch
    aux%y = a - b%y
    call rnd_switch
    aux%z = a - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_single_st = aux
  end function sub_single_st
  !
  type (single_st) function sub_st_single(a,b)
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b
    if (myrandom()) call rnd_switch
    aux%y = a%y - b
    call rnd_switch
    aux%z = a%z - b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_st_single = aux
  end function sub_st_single
  !
  type (single_st) function sub_int_st(a,b)
    implicit none
    integer, intent(in) :: a
    type (single_st), intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a - b%x
    if (myrandom()) call rnd_switch
    aux%y = a - b%y
    call rnd_switch
    aux%z = a - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_int_st = aux
  end function sub_int_st
  !
  type (single_st) function sub_st_int(a,b)
    implicit none
    type (single_st), intent(in) :: a
    integer, intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b
    if (myrandom()) call rnd_switch
    aux%y = a%y - b
    call rnd_switch
    aux%z = a%z - b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_st_int = aux
  end function sub_st_int
  !
  type (double_st) function sub_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a, b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y - b%y
    call rnd_switch
    aux%z = a%z - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_dst_dst = aux
  end function sub_dst_dst
  !
  type (double_st) function sub_dst_double(a,b)
    implicit none
    type (double_st), intent(in) ::a
    double precision, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b
    if (myrandom()) call rnd_switch
    aux%y = a%y - b
    call rnd_switch
    aux%z = a%z - b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_dst_double = aux
  end function sub_dst_double
  !
  type (double_st) function sub_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a - b%x
    if (myrandom()) call rnd_switch
    aux%y = a - b%y
    call rnd_switch
    aux%z = a - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_double_dst = aux
  end function sub_double_dst
  !
  type (double_st) function sub_dst_st(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y - b%y
    call rnd_switch
    aux%z = a%z - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_dst_st = aux
  end function sub_dst_st
  !
  type (double_st) function sub_st_dst(a,b)
    implicit none
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y - b%y
    call rnd_switch
    aux%z = a%z - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_st_dst = aux
  end function sub_st_dst
  !
  type (double_st) function sub_single_dst(a,b)
    implicit none
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a - b%x
    if (myrandom()) call rnd_switch
    aux%y = a - b%y
    call rnd_switch
    aux%z = a - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_single_dst = aux
  end function sub_single_dst
  !
  type (double_st) function sub_dst_single(a,b)
    implicit none
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b
    if (myrandom()) call rnd_switch
    aux%y = a%y - b
    call rnd_switch
    aux%z = a%z - b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_dst_single = aux
  end function sub_dst_single
  !
  type (double_st) function sub_double_st(a,b)
    implicit none
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a - b%x
    if (myrandom()) call rnd_switch
    aux%y = a - b%y
    call rnd_switch
    aux%z = a - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_double_st = aux
  end function sub_double_st
  !
  type (double_st) function sub_st_double(a,b)
    implicit none
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b
    if (myrandom()) call rnd_switch
    aux%y = a%y - b
    call rnd_switch
    aux%z = a%z - b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_st_double = aux
  end function sub_st_double
  !
  type (double_st) function sub_int_dst(a,b)
    implicit none
    integer, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a - b%x
    if (myrandom()) call rnd_switch
    aux%y = a - b%y
    call rnd_switch
    aux%z = a - b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_int_dst = aux
  end function sub_int_dst
  !
  type (double_st) function sub_dst_int(a,b)
    implicit none
    type (double_st), intent(in) :: a
    integer, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x - b
    if (myrandom()) call rnd_switch
    aux%y = a%y - b
    call rnd_switch
    aux%z = a%z - b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    sub_dst_int = aux
  end function sub_dst_int
  !
  !------------------------------------------------------------
  !
  function sub_vdst(a)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: sub_vdst
    sub_vdst%x = - a%x
    sub_vdst%y = - a%y
    sub_vdst%z = - a%z
    sub_vdst%acc = a%acc
  end function sub_vdst
  !
  function sub_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a,b
    type (double_st), dimension(size(a)) :: sub_vdst_vdst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         sub_vdst_vdst(i) = sub_dst_dst(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            sub_vdst_vdst(i)%x = a(i)%x - b(i)%x 
!         else
!            sub_vdst_vdst(i)%x = -((-a(i)%x) - (-b(i)%x))
!         endif
!         if (myrandom()) then
!            sub_vdst_vdst(i)%y = a(i)%y - b(i)%y
!            sub_vdst_vdst(i)%z = -((-a(i)%z) - (-b(i)%z))
!         else
!            sub_vdst_vdst(i)%y = -((-a(i)%y) - (-b(i)%y))
!            sub_vdst_vdst(i)%z = a(i)%z - b(i)%z
!         endif
!         sub_vdst_vdst(i)%acc = -1
!      enddo
!    endif
  end function sub_vdst_vdst
  !
  function sub_dst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: sub_dst_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         sub_dst_vdst(i) = sub_dst_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            sub_dst_vdst(i)%x = a%x - b(i)%x 
!         else
!            sub_dst_vdst(i)%x = -((-a%x) - (-b(i)%x))
!         endif
!         if (myrandom()) then
!            sub_dst_vdst(i)%y = a%y - b(i)%y
!            sub_dst_vdst(i)%z = -((-a%z) - (-b(i)%z))
!         else
!            sub_dst_vdst(i)%y = -((-a%y) - (-b(i)%y))
!            sub_dst_vdst(i)%z = a%z - b(i)%z
!         endif
!         sub_dst_vdst(i)%acc = -1
!      enddo
!    endif
  end function sub_dst_vdst
  !
  function sub_vdst_dst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: sub_vdst_dst
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         sub_vdst_dst(i) = sub_dst_dst(a(i), b)
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           sub_vdst_dst(i)%x = a(i)%x - b%x 
 !        else
 !           sub_vdst_dst(i)%x = -((-a(i)%x) - (-b%x))
 !        endif
 !        if (myrandom()) then
 !           sub_vdst_dst(i)%y = a(i)%y - b%y
 !           sub_vdst_dst(i)%z = -((-a(i)%z) - (-b%z))
 !        else
 !           sub_vdst_dst(i)%y = -((-a(i)%y) - (-b%y))
 !           sub_vdst_dst(i)%z = a(i)%z - b%z
 !        endif
 !        sub_vdst_dst(i)%acc = -1
 !     enddo
 !   endif
  end function sub_vdst_dst
  !
  function sub_double_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b)) :: sub_double_vdst
    integer :: k, i
    k = size(b)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         sub_double_vdst(i) = sub_double_dst(a, b(i))
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           sub_double_vdst(i)%x = a - b(i)%x 
 !        else
 !           sub_double_vdst(i)%x = -((-a) - (-b(i)%x))
 !        endif
 !        if (myrandom()) then
 !           sub_double_vdst(i)%y = a - b(i)%y
 !           sub_double_vdst(i)%z = -((-a) - (-b(i)%z))
 !        else
 !           sub_double_vdst(i)%y = -((-a) - (-b(i)%y))
 !           sub_double_vdst(i)%z = a - b(i)%z
 !        endif
 !        sub_double_vdst(i)%acc = -1
 !     enddo
 !   endif
  end function sub_double_vdst
  !
  function sub_vdst_double(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a)) :: sub_vdst_double
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         sub_vdst_double(i) = sub_dst_double(a(i), b)
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           sub_vdst_double(i)%x = a(i)%x - b
 !        else
 !           sub_vdst_double(i)%x = -((-a(i)%x) - (-b))
 !        endif
 !        if (myrandom()) then
 !           sub_vdst_double(i)%y = a(i)%y - b
 !           sub_vdst_double(i)%z = -((-a(i)%z) - (-b))
 !        else
 !           sub_vdst_double(i)%y = -((-a(i)%y) - (-b))
 !           sub_vdst_double(i)%z = a(i)%z - b
 !        endif
 !        sub_vdst_double(i)%acc = -1
 !     enddo
 !   endif
  end function sub_vdst_double
  !
  function sub_vdouble_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, dimension(:), intent(in) :: a
    type (double_st), dimension(size(b)) :: sub_vdouble_vdst
    integer :: k, i
    k = size(b)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         sub_vdouble_vdst(i) = sub_double_dst(a(i), b(i))
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           sub_vdouble_vdst(i)%x = a(i) - b(i)%x 
 !        else
 !           sub_vdouble_vdst(i)%x = -((-a(i)) - (-b(i)%x))
 !        endif
 !        if (myrandom()) then
 !           sub_vdouble_vdst(i)%y = a(i) - b(i)%y
 !           sub_vdouble_vdst(i)%z = -((-a(i)) - (-b(i)%z))
!         else
!            sub_vdouble_vdst(i)%y = -((-a(i)) - (-b(i)%y))
!            sub_vdouble_vdst(i)%z = a(i) - b(i)%z
!         endif
!         sub_vdouble_vdst(i)%acc = -1
!      enddo
!    endif
  end function sub_vdouble_vdst
  !
  function sub_vdst_vdouble(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(a)) :: sub_vdst_vdouble
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         sub_vdst_vdouble(i) = sub_dst_double(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            sub_vdst_vdouble(i)%x = a(i)%x - b(i) 
!         else
!            sub_vdst_vdouble(i)%x = -((-a(i)%x) - (-b(i)))
!         endif
!         if (myrandom()) then
!            sub_vdst_vdouble(i)%y = a(i)%y - b(i)
!            sub_vdst_vdouble(i)%z = -((-a(i)%z) - (-b(i)))
!         else
!            sub_vdst_vdouble(i)%y = -((-a(i)%y) - (-b(i)))
!            sub_vdst_vdouble(i)%z = a(i)%z - b(i)
!         endif
!         sub_vdst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function sub_vdst_vdouble
  !
  function sub_dst_vdouble(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: sub_dst_vdouble
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         sub_dst_vdouble(i) = sub_dst_double(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            sub_dst_vdouble(i)%x = a%x - b(i)
!         else
!            sub_dst_vdouble(i)%x = -((-a%x) - (-b(i)))
!         endif
!         if (myrandom()) then
!           sub_dst_vdouble(i)%y = a%y - b(i)
!            sub_dst_vdouble(i)%z = -((-a%z) - (-b(i)))
!         else
!            sub_dst_vdouble(i)%y = -((-a%y) - (-b(i)))
!            sub_dst_vdouble(i)%z = a%z - b(i)
!         endif
!         sub_dst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function sub_dst_vdouble
  !
  function sub_vdouble_dst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: sub_vdouble_dst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         sub_vdouble_dst(i) = sub_double_dst(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            sub_vdouble_dst(i)%x = a(i) - b%x 
!         else
!            sub_vdouble_dst(i)%x = -((-a(i)) - (-b%x))
!         endif
!         if (myrandom()) then
!            sub_vdouble_dst(i)%y = a(i) - b%y
!            sub_vdouble_dst(i)%z = -((-a(i)) - (-b%z))
!         else
!            sub_vdouble_dst(i)%y = -((-a(i)) - (-b%y))
!            sub_vdouble_dst(i)%z = a(i) - b%z
!         endif
!         sub_vdouble_dst(i)%acc = -1
!      enddo
!    endif
  end function sub_vdouble_dst
  !
  function sub_vvdst_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a,b
    type (double_st), dimension(size(a,1),size(a,2)) :: sub_vvdst_vvdst
    integer :: i, j, n, m
    n=size(a,1)
    m=size(a,2)
    do j=1,m
       do i=1,n
          sub_vvdst_vvdst(i,j)=sub_dst_dst(a(i,j), b(i,j))
       enddo
    enddo
  end function sub_vvdst_vvdst
  !
  function sub_vvdst(a)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1),size(a,2)) :: sub_vvdst
    integer :: i,j
    do j=1,size(a,2)
       do i=1,size(a,1)
          sub_vvdst(i,j)=-a(i,j)
       enddo
    enddo
  end function sub_vvdst
  !
  function sub_vvvdst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a,b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: sub_vvvdst_vvvdst
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             sub_vvvdst_vvvdst(i,j,k)=sub_dst_dst(a(i,j,k), b(i,j,k))
          enddo
       enddo
    enddo
  end function sub_vvvdst_vvvdst
  !
  function sub_vvvdst_double(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: sub_vvvdst_double
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             sub_vvvdst_double(i,j,k)=sub_dst_double(a(i,j,k), b)
          enddo
       enddo
    enddo
  end function sub_vvvdst_double
  !
  function sub_double_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b,1),size(b,2),size(b,3)) :: sub_double_vvvdst
    integer :: i, j, k
    do k=1,size(b,3)
       do j=1,size(b,2)
          do i=1,size(b,1)
             sub_double_vvvdst(i,j,k)=sub_double_dst(a, b(i,j,k))
          enddo
       enddo
    enddo
  end function sub_double_vvvdst
end module cadna_sub
