!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_type
!  NAME
!    cadna_type
!
!  DESCRIPTION
!    Contains the definitions of the stochastic types, internel variables and functions for
!    managing the numerical debugging and for initializing and closing the CADNA library.
!
!******
module cadna_type
  !      =============
!****s cadna_type/single_st
!  NAME
!    single_st
!
!
!  DESCRIPTION
!    The stochastic type in simple precision.
!
!******
  type single_st
     sequence
     real :: x,y,z
!     integer :: acc = -1
     integer :: acc 
  end type single_st
  !
!****s cadna_type/double_st
!  NAME
!    double_st
!
!
!  DESCRIPTION
!    The stochastic type in double precision.
!
!******
  type double_st
     sequence
     real(kind(0.d0)) :: x,y,z
!     integer :: acc = -1
     integer :: acc 
     integer :: pad
  end type double_st
  !
!****d 
!  NAME
!    cadna_"name-of-instability"
!
!
!  DESCRIPTION
!    Internal integer parameters to specify a kind of 
!    numerical instability
!
!******
  integer, parameter :: &
       & cadna_branching = 8, &
       & cadna_mathematic = 16, &
       & cadna_intrinsic = 32, &
       & cadna_cancellation = 64, &
       & cadna_division = 128, &
       & cadna_power = 256, &
       & cadna_multiplication = 512, &
       & cadna_all = 2048
  !
!****d 
!  NAME
!    cadna_"name-of-instability"_tag
!
!
!  DESCRIPTION
!    Internal logical parameters to specify if the detection of a kind 
!    of numerical instability is enabled or not
!
!******
  logical :: &
       & cadna_division_tag = .true., &
       & cadna_power_tag = .true., &
       & cadna_multiplication_tag = .true., &
       & cadna_branching_tag = .true., &
       & cadna_mathematic_tag = .true., &
       & cadna_intrinsic_tag = .true., &
       & cadna_cancellation_tag = .true., &
       & deactivate = .false.
  !
!****d 
!  NAME
!    cadna_"name-of-instability"_count
!
!  DESCRIPTION
!    Internal integer parameters to count the number of numerical 
!    instabilities that have been detected for each kind of instability.
!
!******
  integer :: &
       & cadna_division_count = 0, &
       & cadna_power_count = 0, &
       & cadna_multiplication_count = 0, &
       & cadna_branching_count = 0, &
       & cadna_mathematic_count = 0, &
       & cadna_intrinsic_count = 0, &
       & cadna_cancellation_count=0
  !
!****d 
!  NAME
!    cadna_"name-of-instability"_msg
!
!  DESCRIPTION
!    Internal string parameters containing the name of each 
!    numerical instability.
!
!******
  character(LEN=34) :: &
       & cadna_division_msg = 'UNSTABLE DIVISION(S)', &
       & cadna_power_msg = 'UNSTABLE POWER FUNCTION(S)', &
       & cadna_multiplication_msg = 'UNSTABLE MULTIPLICATION(S)', &
       & cadna_branching_msg = 'UNSTABLE BRANCHING(S)', &
       & cadna_mathematic_msg = 'UNSTABLE MATHEMATICAL FUNCTION(S)', &
       & cadna_intrinsic_msg = 'UNSTABLE INTRINSIC FUNCTION(S)', &
       & cadna_cancellation_msg = 'UNSTABLE CANCELLATION(S)'
  !
!****d 
!  DESCRIPTION
!    Internal integer parameters to manage the detection of 
!    numerical instabilities.
!
!******
  integer :: cadna_numb_error_max, cadna_numb_error, cadna_cancellation_level = 4
  !
!****d 
!  DESCRIPTION
!    Internal integer parameters for the random number generator.
!
!******
  integer :: recurrence, irandom = 28
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
!****f cadna_type/cadna_enable
!  NAME
!    cadna_enable
!
!  FUNCTION
!    Subroutine to enable the detection of a kind of
!    numerical instability.
!
!  INPUTS
!    cadna_parameter : integer 
!
!******
  subroutine cadna_enable(cadna_parameter)
    integer, intent(in) :: cadna_parameter
    select case (cadna_parameter)
       case (cadna_branching)
          cadna_branching_tag = .true.
       case (cadna_mathematic)
          cadna_mathematic_tag = .true.
       case (cadna_intrinsic)
          cadna_intrinsic_tag = .true.
       case (cadna_cancellation)
          cadna_cancellation_tag = .true.
       case (cadna_power)
          cadna_power_tag = .true.
       case (cadna_multiplication)
          cadna_multiplication_tag = .true.
       case (cadna_division)
          cadna_division_tag = .true.
       end select
  end subroutine cadna_enable
  !
!****f cadna_type/cadna_disable
!  NAME
!    cadna_disable
!
!  FUNCTION
!    Subroutine to disable the detection of a kind of
!    numerical instability.
!
!  INPUTS
!    cadna_parameter: integer 
!
!******
  subroutine cadna_disable(cadna_parameter)
    integer, intent(in) :: cadna_parameter
    select case (cadna_parameter)
       case (cadna_branching)
          cadna_branching_tag = .false.
       case (cadna_mathematic)
          cadna_mathematic_tag = .false.
       case (cadna_intrinsic)
          cadna_intrinsic_tag = .false.
       case (cadna_cancellation)
          cadna_cancellation_tag = .false.
       case (cadna_power)
          cadna_power_tag = .false.
          deactivate = .true.
       case (cadna_multiplication)
          cadna_multiplication_tag = .false.
          deactivate = .true.
       case (cadna_division)
          cadna_division_tag = .false.
          deactivate = .true.
       end select
  end subroutine cadna_disable
  !
!****f cadna_type/cadna_init
!  NAME
!    cadna_init
!
!  FUNCTION
!    Subroutine for initialising the CADNA library.
!
!  INPUTS
!    numb_instability : integer,
!    to specify the number of instability that will be detected, if numb_instability = -1, 
!    there is no limit for the dectection.
!
!    cadna_instability : integer,
!    An optional argument to specify what kinds of instability will be DISABLED !!
!    By default, every kind of instability is enabled.
!
!    cancel_level : integer,
!    An optional argument to specify the level from which the "CANCELLATION" instability
!    is pointed out. The default value is 4.
!
!    init_random : integer, 
!    An optional argument to initialise the random number generator. The default value is 28.
!
!******
  subroutine cadna_init(numb_instability, cadna_instability, cancel_level, init_random)
    implicit none
    integer, optional :: cadna_instability
    integer, optional :: cancel_level
    integer, optional :: init_random
    integer :: numb_instability
    character(3) :: caux
    !
    if (present(init_random)) then
       recurrence = init_random
    else
       recurrence = 51
    endif
    !
    if (present(cadna_instability)) then
       if (cadna_instability == cadna_all) then
          cadna_division_tag = .false.
          cadna_power_tag = .false.
          cadna_multiplication_tag = .false.
          cadna_branching_tag = .false.
          cadna_mathematic_tag = .false.
          cadna_intrinsic_tag = .false.
          cadna_cancellation_tag = .false.
          deactivate = .true.
       else
          cadna_branching_tag = mod(cadna_instability/cadna_branching,2) == 0
          cadna_mathematic_tag = mod(cadna_instability/cadna_mathematic,2) == 0
          cadna_intrinsic_tag = mod(cadna_instability/cadna_intrinsic,2) == 0
          cadna_cancellation_tag = mod(cadna_instability/cadna_cancellation,2) == 0
          cadna_division_tag = mod(cadna_instability/cadna_division,2) == 0
          cadna_power_tag = mod(cadna_instability/cadna_power,2) == 0
          cadna_multiplication_tag = mod(cadna_instability/cadna_multiplication,2) == 0
       endif
       if (cadna_power_tag.and.cadna_division_tag.and.cadna_multiplication_tag) then
       else
          deactivate = .true.
       endif
    endif
    !
    if (present(cancel_level)) then
       cadna_cancellation_level = cancel_level
    endif
    !
    cadna_numb_error_max = numb_instability
    cadna_numb_error = 0
    if (cadna_numb_error_max == 0) then
       cadna_division_tag = .false.
       cadna_power_tag = .false.
       cadna_multiplication_tag = .false.
       deactivate = .true.
       cadna_branching_tag = .false.
       cadna_mathematic_tag = .false.
       cadna_intrinsic_tag = .false.
       cadna_cancellation_tag = .false.
       deactivate = .true.
    end if
    !
    write(*,*) "-----------------------------------------------------"
    write(*,*) "CADNA software --- University P. et M. Curie --- LIP6"
    if (deactivate) then
       caux = "OFF"
    else
       caux = "ON "
    endif
    write(*,*) "Self-validation detection: ", caux
    if (cadna_mathematic_tag) then
       caux = "ON "
    else
       caux = "OFF"
    endif
    write(*,*) "Mathematical instabilities detection: ", caux
    if (cadna_branching_tag) then
       caux = "ON "
    else
       caux = "OFF"
    endif
    write(*,*) "Branching instabilities detection: ", caux
    if (cadna_intrinsic_tag) then
       caux = "ON "
    else
       caux = "OFF"
    endif
    write(*,*) "Intrinsic instabilities detection: ", caux
    if (cadna_cancellation_tag) then
       caux = "ON "
    else
       caux = "OFF"
    endif
    write(*,*) "Cancellation instabilities detection: ", caux
    write(*,*) "-----------------------------------------------------"
    call rnd_moinf
  end subroutine cadna_init
  !
!****if cadna_type/myrandom
!  NAME
!    myrandom
!
!  FUNCTION
!    An internal random number generator.
!
!******
  logical function myrandom()
    integer :: ia, im, ip0, ip1
    parameter (ia = 31415821, im = 100000000)
    myrandom = btest(recurrence,irandom)
    irandom = irandom - 1
    if (irandom.eq.2) then
       irandom = 28
       ip0 = mod(recurrence,10000)
       ip1 = recurrence/10000
       recurrence = mod(ip0*3141+ip1*5821,10000)*10000
       recurrence = mod(recurrence+ip0*5821+1,100000000)
    endif
  end function myrandom
  !
!****if cadna_type/cadna_end
!  NAME
!    cadna_end
!
!  FUNCTION
!    Subroutine for closing the CADNA library and for printing out
!    informations on instabilities that have beeen detected.
!
!******
  subroutine cadna_end()
    write(*,*) "-----------------------------------------------------"
    write(*,*) "CADNA software --- University P. et M. Curie --- LIP6"
    if (deactivate) write(*,*)"BE CAREFUL: the self-validation has been disconnected"
    if ( cadna_numb_error .eq. 0) then 
       write(*,*) "No instability detected"
    else
       if (cadna_division_count+cadna_power_count+cadna_multiplication_count.ne.0) then
          write(*,*) "CRITICAL WARNING: the self-validation detects major problem(s)."
          write(*,*) "The results are NOT guaranteed"
       endif
       if (cadna_numb_error.eq.1) then
          write(*,*) "There is  1 numerical instability"
       else 
          write(*,*) "There are ",cadna_numb_error," numerical instabilities"
       endif
       write(*,*) cadna_division_count, " ", cadna_division_msg
       write(*,*) cadna_power_count, " ", cadna_power_msg
       write(*,*) cadna_multiplication_count, " ", cadna_multiplication_msg
       write(*,*) cadna_branching_count, " ", cadna_branching_msg
       write(*,*) cadna_mathematic_count, " ", cadna_mathematic_msg
       write(*,*) cadna_intrinsic_count, " ", cadna_intrinsic_msg
       write(*,*) cadna_cancellation_count, " ", cadna_cancellation_msg
    endif
    write(*,*) "-----------------------------------------------------"
  end subroutine cadna_end
  !
end module cadna_type
